from __future__ import unicode_literals

from django import forms
from django.db import models
from django.contrib.contenttypes.fields import GenericRelation
from django.utils.encoding import smart_text

from pagetree.models import PageBlock


class BasePageBlock(models.Model):
    """An abstract pageblock to be used for custom pageblocks."""

    class Meta:
        abstract = True

    pageblocks = GenericRelation(PageBlock)

    display_name = None
    """The name for this block that's used in pagetree's menus."""

    def as_dict(self):
        """Returns the PageBlock as a dictionary.

        :rtype: dictionary
        """
        return dict(block_type=self.display_name)

    def allow_redo(self):
        """Determines whether this block allows users to resubmit answers.

        This method is only used if ``needs_submit`` is True. If
        ``allow_redo`` is also True, pagetree will render a button
        with the text "I want to change my answers." when the block
        has been submitted.

        :returns: a boolean
        """

        return False

    def needs_submit(self):
        """Determines whether this pageblock needs form controls rendered.

        If ``needs_submit`` is True, then pagetree will create a <form>
        on this pageblock's surrounding page, and a Submit button for
        that form. It may also render a "Clear results" button, under
        the right circumstances. The surrounding <form> allows pagetree
        to handle form submissions for multiple blocks on the same page.

        Also, when ``needs_submit`` is True, the POST data on the
        Section's ``submit()`` step gets processed, but when needs_submit
        is False, nothing is sent to the server.

        :returns: a boolean
        """

        return False

    def unlocked(self, user):
        """Determines whether the user can proceed past this block.

        The current user is passed in to this function, allowing you to,
        for example, find out if that user has submitted the info
        necessary to proceed past this block's page.

        :param user: the current user
        :returns: a boolean
        """

        return True

    def submit(self, user, request_data):
        """This is a hook for handling this pageblock's form submission.

        :returns: None
        """

        pass

    def redirect_to_self_on_submit(self):
        """Determines where the user is redirected to on page submission.

        If ``redirect_to_self_on_submit`` returns True, then the user will
        be redirected to the current page when the page's form is
        submitted. Otherwise, they'll be redirected to the next page in
        the hierarchy.

        :returns: a boolean
        """

        return True

    def clear_user_submissions(self, user):
        """A hook to clear any user submissions for this block.

        :returns: None
        """

        pass

    def pageblock(self):
        return self.pageblocks.first()

    def __str__(self):
        return smart_text(self.pageblock())

    # TODO: I'd like to have all the following methods be inherited
    # somehow. For now these need to be copy and pasted for each custom
    # pageblock.
    @staticmethod
    def add_form():
        return BasePageBlockForm()

    def edit_form(self):
        return BasePageBlockForm(instance=self)

    @staticmethod
    def create(request):
        form = BasePageBlockForm(request.POST)
        return form.save()

    @classmethod
    def create_from_dict(cls, d):
        return cls.objects.create(**d)

    def edit(self, vals, files):
        form = BasePageBlockForm(data=vals, files=files, instance=self)
        if form.is_valid():
            form.save()


class BasePageBlockForm(forms.ModelForm):
    """Example ModelForm for the BasePageBlock.

    This is just an example. It should always be replaced with your
    own ModelForm pointing to the custom pageblock.
    """

    class Meta:
        model = BasePageBlock
        fields = '__all__'



from __future__ import unicode_literals

from django.conf.urls import include, url
from pagetree.generic.views import EditView, InstructorView, PageView

urlpatterns = [
    url(r'^pagetree/', include('pagetree.urls')),

    # instantiate generic views using a template that doesn't
    # extend/depend on base.html
    url(r'^pages/edit/(?P<path>.*)$',
        EditView.as_view(template_name="pagetree/test_page.html"),
        {}, 'edit-page'),
    url(r'^pages/instructor/(?P<path>.*)$',
        InstructorView.as_view(template_name="pagetree/test_page.html")),
    url(r'^pages/(?P<path>.*)$',
        PageView.as_view(template_name="pagetree/test_page.html")),

    # a second set to make sure non-default hierarchies
    # get handled as well
    url(r'^pages2/edit/(?P<path>.*)$',
        EditView.as_view(
            hierarchy_name="two",
            hierarchy_base="/pages2",
            template_name="pagetree/test_page.html"),
        {}, 'edit-page'),
    url(r'^pages2/instructor/(?P<path>.*)$',
        InstructorView.as_view(
            hierarchy_name="two",
            hierarchy_base="/pages2",
            template_name="pagetree/test_page.html")),
    url(r'^pages2/(?P<path>.*)$',
        PageView.as_view(
            hierarchy_name="two",
            hierarchy_base="/pages2",
            template_name="pagetree/test_page.html")),
]



from __future__ import unicode_literals

from django import forms
from django.db import models
from django.utils.encoding import smart_text

from pagetree.generic.models import BasePageBlock
from pagetree.reports import ReportableInterface, ReportColumnInterface


class TestReportColumn(ReportColumnInterface):

    def identifier(self):
        return "Test Report Column"

    def metadata(self):
        return ['', self.identifier(), "generic",
                "string", "this is a test"]

    def user_value(self, user):
        return user.username


class TestBlock(BasePageBlock):
    """ this is a pageblock that is exclusively for pagetree's
    internal tests so we have some kind of block to test with
    without having to pull in the whole django-pageblocks
    package and its dependencies.

    You should never use it. In fact, forget that you ever saw this.
    """

    class Meta:
        app_label = 'pagetree'

    body = models.TextField(blank=True)

    template_file = "pagetree/testblock.html"
    display_name = "Test Block"

    def __str__(self):
        return smart_text(self.pageblock())

    def pageblock(self):
        return self.pageblocks.first()

    @classmethod
    def add_form(cls):
        class AddForm(forms.Form):
            body = forms.CharField(
                widget=forms.widgets.Textarea(attrs={'cols': 80}))
        return AddForm()

    @classmethod
    def create(cls, request):
        return TestBlock.objects.create(body=request.POST.get('body', ''))

    @classmethod
    def create_from_dict(cls, d):
        return cls.objects.create(**d)

    def edit_form(self):
        class EditForm(forms.Form):
            body = forms.CharField(widget=forms.widgets.Textarea(),
                                   initial=self.body)
        return EditForm()

    def edit(self, vals, files):
        self.body = vals.get('body', '')
        self.save()

    def as_dict(self):
        d = super(TestBlock, self).as_dict()
        d['body'] = self.body
        return d

    def import_from_dict(self, d):
        self.body = d.get('body', '')
        self.save()

    def summary_render(self):
        if len(self.body) < 61:
            return self.body
        else:
            return self.body[:61] + "..."

    def report_metadata(self):
        return [TestReportColumn()]

    def report_values(self):
        return [TestReportColumn()]


ReportableInterface.register(TestBlock)



from __future__ import unicode_literals

from pagetree.models import Hierarchy
from pagetree.reports import StandaloneReportColumn, PagetreeReport
from pagetree.tests.factories import PagetreeTestCase, UserFactory


class PagetreeReportColumnTest(PagetreeTestCase):

    def setUp(self):
        super(PagetreeReportColumnTest, self).setUp()

        self.column = StandaloneReportColumn('username', 'profile',
                                             'string', 'Username',
                                             lambda x: x.username)

    def test_identifier(self):
        self.assertEquals(self.column.identifier(), 'username')

    def test_metadata(self):
        key_row = ['', 'username', 'profile', 'string', 'Username']
        self.assertEquals(self.column.metadata(), key_row)

    def test_user_value(self):
        user = UserFactory()
        self.assertEquals(self.column.user_value(user), user.username)


class PagetreeReportTest(PagetreeTestCase):

    def setUp(self):
        super(PagetreeReportTest, self).setUp()
        self.report = PagetreeReport()
        self.user = UserFactory()

    def test_users(self):
        self.assertEquals(len(self.report.users()), 1)

    def test_standalone_columns(self):
        self.assertEquals(len(self.report.standalone_columns()), 0)

    def test_metadata_columns(self):
        hierarchies = Hierarchy.objects.all()
        self.assertEquals(len(hierarchies), 2)
        columns = self.report.metadata_columns(hierarchies)
        self.assertEquals(len(columns), 2)

    def test_metadata(self):
        hierarchies = Hierarchy.objects.all()
        rows = []
        for row in self.report.metadata(hierarchies):
            rows.append(row)
        self.assertEquals(len(rows), 4)

    def test_value_columns(self):
        hierarchies = Hierarchy.objects.all()
        columns = self.report.value_columns(hierarchies)
        self.assertEquals(len(columns), 2)

    def test_values(self):
        hierarchies = Hierarchy.objects.all()
        rows = []
        for row in self.report.values(hierarchies):
            rows.append(row)
        self.assertEquals(len(rows), 2)

        self.assertEquals(rows[1][0], self.user.username)



from __future__ import unicode_literals

from django.test import TestCase
from django.contrib.auth.models import User
from django.http import Http404
from django.utils.encoding import smart_text
import django.db
from django.db import transaction

from pagetree.models import Hierarchy, PageBlock, UserPageVisit
from pagetree.test_models import TestBlock
from pagetree.tests.factories import (
    RootSectionFactory, TestBlockFactory, UserFactory, UserPageVisitFactory,
    HierarchyFactory
)


class UserTest(TestCase):
    def setUp(self):
        self.u = UserFactory()

    def test_is_valid_from_factory(self):
        self.u.full_clean()


class HierarchyTest(TestCase):
    def setUp(self):
        self.h = HierarchyFactory()

    def test_is_valid_from_factory(self):
        self.h.full_clean()


class RootSectionTest(TestCase):
    def setUp(self):
        self.s = RootSectionFactory()

    def test_is_valid_from_factory(self):
        self.s.full_clean()


class EmptyHierarchyTest(TestCase):
    """ a hierarchy with no sections in it
    (one Root gets created by default) """
    def setUp(self):
        self.h = Hierarchy.from_dict({'name': "main",
                                      'base_url': ""})

    def tearDown(self):
        self.h.delete()

    def test_easy(self):
        self.assertEqual(self.h.name, "main")

    def test_absolute_url(self):
        self.assertEqual(self.h.get_absolute_url(), "")

    def test_empty_hierarchy(self):
        self.assertEqual(self.h.get_root().hierarchy.id, self.h.id)

    def test_invalid_path(self):
        self.assertEqual(self.h.find_section_from_path("/foo/bar/baz/"),
                         None)

    def test_valid_path(self):
        self.assertEqual(self.h.find_section_from_path("/"),
                         self.h.get_root())

    def test_get_section_from_path(self):
        self.assertEqual(self.h.get_section_from_path("/"),
                         self.h.get_root())

    def test_get_section_from_path_invalid(self):
        with self.assertRaises(Http404):
            self.h.get_section_from_path("/foo/bar/baz")

    def test_get_first_leaf(self):
        self.assertEqual(
            self.h.get_first_leaf(self.h.get_root()).id,
            self.h.get_root().id)

    def test_get_last_leaf(self):
        self.assertEqual(
            self.h.get_last_leaf(self.h.get_root()).id,
            self.h.get_root().id)

    def test_as_dict(self):
        self.assertEqual(
            'name' in self.h.as_dict(),
            True)
        self.assertEqual(
            'base_url' in self.h.as_dict(),
            True)
        self.assertEqual(
            'sections' in self.h.as_dict(),
            True)

    def test_get_hierarchy(self):
        self.assertEqual(
            self.h.id,
            Hierarchy.get_hierarchy("main").id)

    def test_str(self):
        self.assertEqual(smart_text(self.h), "main")

    def test_empty_top_level(self):
        self.assertEqual(
            len(self.h.get_top_level()),
            0)

    def test_available_pageblocks_empty(self):
        with self.settings(PAGEBLOCKS=None):
            self.assertEqual(self.h.available_pageblocks(), [])

    def test_add_section_from_dict(self):
        s = self.h.add_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [
                    {
                        'label': 'Section 2',
                        'slug': 'section-2',
                        'pageblocks': [],
                        'children': [],
                    }],
            })
        # adding straight to the hierarchy needs to
        # create a Root section, not a regular one
        self.assertEqual(s.label, 'Root')
        self.assertEqual(s.slug, '')


class OneLevelDeepTest(TestCase):
    def setUp(self):
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'section-2',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 3',
                'slug': 'section-3',
                'pageblocks': [],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.section2 = r[1]
        self.section3 = r[2]

    def tearDown(self):
        self.h.delete()

    def test_root(self):
        self.assertEqual(
            self.h.get_root().id,
            self.root.id)

    def test_children(self):
        self.assertEqual(
            self.section1.label,
            "Section 1")

    def test_add_child_form(self):
        f = self.section1.add_child_section_form()
        self.assertEqual(
            'label' in f.fields,
            True)

    def test_edit_form(self):
        f = self.section1.edit_form()
        self.assertEqual(
            'label' in f.fields,
            True)
        self.assertEqual(
            'slug' in f.fields,
            True)

    def test_get_path(self):
        self.assertEqual(
            self.section1.get_path(),
            "section-1/")

    def test_add_pageblock_form(self):
        f = self.section1.add_pageblock_form()
        self.assertEqual(
            'label' in f.fields,
            True)
        self.assertEqual(
            'css_extra' in f.fields,
            True)
        self.assertFalse(f.fields['label'].required)
        self.assertFalse(f.fields['css_extra'].required)

    def test_valid_path(self):
        self.assertEqual(self.h.find_section_from_path("section-1/"),
                         self.section1)

    def test_trailing_slash_ignored(self):
        self.assertEqual(self.h.find_section_from_path("section-1/"),
                         self.h.find_section_from_path("section-1"))

    def test_get_first_leaf(self):
        self.assertEqual(self.h.get_first_leaf(self.root),
                         self.section1)

    def test_get_last_leaf(self):
        self.assertEqual(self.h.get_last_leaf(self.root),
                         self.section3)

    def test_get_module(self):
        self.assertEqual(
            self.root.get_module(),
            None)
        self.assertEqual(
            self.section1.get_module(),
            self.section1)

    def test_is_first_child(self):
        self.assertEqual(
            self.root.is_first_child(),
            True)
        self.assertEqual(
            self.section1.is_first_child(),
            True)
        self.assertEqual(
            self.section2.is_first_child(),
            False)

    def test_is_last_child(self):
        self.assertEqual(
            self.root.is_last_child(),
            True)
        self.assertEqual(
            self.section1.is_last_child(),
            False)
        self.assertEqual(
            self.section2.is_last_child(),
            False)
        self.assertEqual(
            self.section3.is_last_child(),
            True)

    def test_get_previous(self):
        self.assertEqual(
            self.section1.get_previous(),
            None)
        self.assertEqual(
            self.section2.get_previous(),
            self.section1)
        self.assertEqual(
            self.section3.get_previous(),
            self.section2)
        self.assertEqual(
            self.root.get_previous(),
            None)

    def test_get_next(self):
        self.assertEqual(
            self.section1.get_next(),
            self.section2)
        self.assertEqual(
            self.section2.get_next(),
            self.section3)
        self.assertEqual(
            self.section3.get_next(),
            None)
        self.assertEqual(
            self.root.get_previous(),
            None)

    def test_section_str(self):
        self.assertEqual(
            smart_text(self.section1),
            "Section 1")

    def test_section_get_absolute_url(self):
        self.assertEqual(
            self.section1.get_absolute_url(),
            "section-1/")

    def test_section_get_edit_url(self):
        self.assertEqual(
            self.section1.get_edit_url(),
            "edit/section-1/")

    def test_section_get_instructor_url(self):
        self.assertEqual(
            self.section1.get_instructor_url(),
            "instructor/section-1/")

    def test_section_get_first_leaf(self):
        self.assertEqual(
            self.section1.get_first_leaf(),
            self.section1)
        self.assertEqual(
            self.root.get_first_leaf(),
            self.section1)

    def test_section_get_last_leaf(self):
        self.assertEqual(
            self.section3.get_last_leaf(),
            self.section3)
        self.assertEqual(
            self.root.get_last_leaf(),
            self.section3)

    def test_update_children_order(self):
        normal_order = [
            self.section1.id,
            self.section2.id,
            self.section3.id]
        normal_order.reverse()
        self.root.update_children_order(normal_order)
        r = self.root.get_children()
        self.assertEqual([s.id for s in r], normal_order)

    def test_get_tree_depth(self):
        self.assertEqual(self.section1.get_tree_depth(), 1)
        self.assertEqual(self.section2.get_tree_depth(), 2)
        self.assertEqual(self.section3.get_tree_depth(), 3)


class OneLevelWithBlocksTest(TestCase):
    def setUp(self):
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [
                    {'label': '',
                     'css_extra': '',
                     'block_type': 'Test Block',
                     'body': 'some body text section 1 block 1',
                     },
                    {'label': '',
                     'css_extra': '',
                     'block_type': 'Test Block',
                     'body': 'some body text section 1 block 2',
                     }
                ],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]

    def tearDown(self):
        self.h.delete()

    def test_str(self):
        b = self.section1.pageblock_set.all()[0]
        self.assertEqual(
            smart_text(b),
            "Section 1 [1]: ")

    def test_edit_label(self):
        b = self.section1.pageblock_set.all()[0]
        self.assertEqual(
            b.edit_label(), 'Test Block')

    def test_render(self):
        b = self.section1.pageblock_set.all()[0]
        self.assertEqual(
            b.render().strip(),
            "<p>some body text section 1 block 1</p>")

    def test_render_js(self):
        b = self.section1.pageblock_set.all()[0]
        self.assertEqual(
            b.render_js().strip(),
            "")

    def test_render_css(self):
        b = self.section1.pageblock_set.all()[0]
        self.assertEqual(
            b.render_css().strip(),
            "")

    def test_render_summary(self):
        b = self.section1.pageblock_set.all()[0]
        self.assertEqual(
            b.render_summary().strip(),
            "some body text section 1 block 1")

    def test_edit_form(self):
        b = self.section1.pageblock_set.all()[0]
        f = b.edit_form()
        self.assertEqual(
            "body" in f.fields,
            True)

    def test_default_edit_form(self):
        b = self.section1.pageblock_set.all()[0]
        f = b.default_edit_form()
        self.assertEqual(
            "label" in f.fields,
            True)
        self.assertEqual(
            "css_extra" in f.fields,
            True)
        self.assertFalse(f.fields['label'].required)
        self.assertFalse(f.fields['css_extra'].required)

    def test_edit(self):
        b = self.section1.pageblock_set.all()[0]
        b.edit(dict(label="new label",
                    css_extra="new css_extra",
                    body="new_body"), None)
        self.assertEqual(b.label, "new label")
        self.assertEqual(b.css_extra, "new css_extra")
        self.assertEqual(b.block().body, "new_body")

    def test_serialization(self):
        self.assertEqual(
            self.section1.as_dict()['pageblocks'][0]['body'],
            'some body text section 1 block 1')

    def test_delete_block(self):
        block1 = self.section1.pageblock_set.all()[0]
        block2 = self.section1.pageblock_set.all()[1]
        block1.delete()
        # block2 should now be #1, but we need to re-fetch it
        block2 = PageBlock.objects.get(id=block2.id)
        self.assertEqual(block2.ordinality, 1)

    def test_closing_children(self):
        self.assertEqual(
            list(self.section1.closing_children()),
            [self.section1])

    def test_move_form(self):
        f = self.section1.move_form()
        self.assertTrue(f is not None)

    def test_update_pageblocks_order(self):
        normal_order = [p.id for p in self.section1.pageblock_set.all()]
        normal_order.reverse()
        self.section1.update_pageblocks_order(normal_order)
        self.assertEqual(
            [p.id for p in self.section1.pageblock_set.all()],
            normal_order)

    def test_import_from_dict(self):
        b = self.section1.pageblock_set.first()
        d = {'label': 'new label',
             'css_extra': 'new css_extra',
             'body': 'new body'}
        b.import_from_dict(d)
        self.assertEquals(b.label, 'new label')
        self.assertEquals(b.css_extra, 'new css_extra')

        sub = b.block()
        self.assertEquals(sub.body, 'new body')

    def test_import_custom_block_dict(self):
        b = self.section1.pageblock_set.first()
        d = {
            'block_type': 'Test Block',
            'body': 'abc',
        }
        b.import_from_dict(d)

        sub = b.block()
        self.assertEquals(sub.display_name, 'Test Block')
        self.assertEquals(sub.body, 'abc')


class UserTrackingTest(TestCase):
    def setUp(self):
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'section-2',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 3',
                'slug': 'section-3',
                'pageblocks': [],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.section2 = r[1]
        self.section3 = r[2]

        self.h2 = Hierarchy.objects.create(name="other", base_url="")
        self.h2.get_root().add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })
        self.h2.get_root().add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })

        self.user = User.objects.create(username='testuser')

    def tearDown(self):
        self.h.delete()
        self.h2.delete()
        self.user.delete()

    def test_user_visit(self):
        self.h.user_visit(self.user, self.section1)
        self.assertEqual(
            self.h.get_user_location(self.user),
            "section-1/"
        )
        self.assertEqual(
            self.h.get_user_section(self.user),
            self.section1
        )

    def test_user_visit_section(self):
        self.section1.user_visit(self.user)
        self.assertEqual(
            self.h.get_user_location(self.user),
            "section-1/"
        )
        self.assertEqual(
            self.h.get_user_section(self.user),
            self.section1
        )

    def test_user_pagevisit(self):
        self.assertEqual(
            self.section1.get_uservisit(self.user),
            None
        )
        self.section1.user_pagevisit(self.user, status="incomplete")
        self.assertEqual(
            self.section1.get_uservisit(self.user).status,
            "incomplete"
        )
        self.section1.user_pagevisit(self.user, status="complete")
        self.assertEqual(
            self.section1.get_uservisit(self.user).status,
            "complete"
        )

    def test_user_pagevisit_multiple(self):
        self.section1.user_pagevisit(self.user, status="complete")
        try:
            with transaction.atomic():
                # then stuff another one in manually to simulate a
                # race condition
                UserPageVisit.objects.create(section=self.section1,
                                             user=self.user,
                                             status="bad status")
                # should not be able to happen
                self.assertEqual(True, False)
        except django.db.IntegrityError:
            self.assertEqual(True, True)
        self.section1.user_pagevisit(self.user, status="complete")

    def test_gate_check(self):
        self.assertEqual(self.section1.gate_check(None),
                         (False, self.section1))
        self.assertEqual(self.section1.gate_check(self.user), (True, None))
        self.assertEqual(self.section2.gate_check(self.user),
                         (False, self.section1))
        self.section1.user_pagevisit(self.user, status="complete")
        self.section2.user_pagevisit(self.user, status="complete")
        self.section3.user_pagevisit(self.user, status="complete")
        self.assertEqual(self.section1.gate_check(self.user), (True, None))

    def test_gate_check_multiple_hierarchies(self):
        children = self.h2.get_root().get_children()

        # first section of any hierarchu returns "True" from gate_check
        self.assertEqual(
            children[0].gate_check(self.user), (True, None))
        self.assertEqual(
            children[1].gate_check(self.user), (False, children[0]))


class VersionTest(TestCase):
    def setUp(self):
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [
                    {'label': '',
                     'css_extra': '',
                     'block_type': 'Test Block',
                     'body': 'some body text section 1 block 1',
                     },
                    {'label': '',
                     'css_extra': '',
                     'block_type': 'Test Block',
                     'body': 'some body text section 1 block 2',
                     }
                ],
                'children': [
                    {
                        'label': 'Child 1',
                        'slug': 'child-1',
                        'pageblocks': [],
                        'children': [{
                                'label': 'GrandChild 1',
                                'slug': 'grandchild-1',
                                'pageblocks': [],
                                'children': [],
                        }],
                    }
                ],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.user = User.objects.create(username='testuser')

    def tearDown(self):
        self.h.delete()
        self.user.delete()

    def test_save_version(self):
        self.assertEqual(
            self.section1.version_set.count(),
            0)
        self.section1.save_version(self.user, activity="test save")
        self.assertEqual(
            self.section1.version_set.count(),
            1)
        v = self.section1.version_set.all()[0]
        self.assertEqual(
            v.activity, "test save")
        self.assertEqual(
            v.user, self.user)
        # just some quick checks that our block bodies made it into
        # the data
        self.assertEqual(
            "some body text section 1 block 1" in v.data,
            True)
        self.assertEqual(
            "some body text section 1 block 2" in v.data,
            True)
        # and make sure the entire subtree got serialized
        self.assertEqual(
            'GrandChild 1' in v.data,
            True)

    def test_more_recent_versions(self):
        self.section1.save_version(self.user, activity="test save")
        # grab the most recent version
        v = list(self.section1.version_set.all())[-1]
        # should not be any newer ones
        self.assertEqual(
            len(v.more_recent_versions()),
            0)
        # add another
        self.section1.save_version(self.user, activity="another test save")
        # now there should be
        self.assertEqual(
            len(v.more_recent_versions()),
            1)

    def test_more_recent_versions_with_children(self):
        self.section1.save_version(self.user, activity="test save")
        # grab the most recent version
        v = list(self.section1.version_set.all())[-1]
        # should not be any newer ones
        self.assertEqual(
            len(v.more_recent_versions()),
            0)
        # add another on a child
        self.section1.get_children()[0].save_version(
            self.user,
            activity="another test save")
        # now there should be
        self.assertEqual(
            len(v.more_recent_versions()),
            1)


class MultipleLevelsTest(TestCase):
    def setUp(self):
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'section-2',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 3',
                'slug': 'section-3',
                'pageblocks': [],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.section2 = r[1]
        self.section3 = r[2]
        self.section3.add_child_section_from_dict(
            {
                'label': 'Section 4',
                'slug': 'section-4',
                'pageblocks': [],
                'children': [],
            })
        self.section4 = self.section3.get_children()[0]
        self.section4.add_child_section_from_dict(
            {
                'label': 'Section 5',
                'slug': 'section-5',
                'pageblocks': [],
                'children': [],
            })
        self.section5 = self.section4.get_children()[0]

    def tearDown(self):
        Hierarchy.objects.all().delete()

    def test_get_absolute_url(self):
        self.assertEqual(self.section4.get_absolute_url(),
                         "section-3/section-4/")

    def test_get_edit_url(self):
        self.assertEqual(self.section4.get_edit_url(),
                         "edit/section-3/section-4/")

    def test_get_absolute_url_two_down(self):
        self.assertEqual(self.section5.get_absolute_url(),
                         "section-3/section-4/section-5/")

    def test_get_edit_url_two_down(self):
        self.assertEqual(self.section5.get_edit_url(),
                         "edit/section-3/section-4/section-5/")

    def test_clone_hierarchy(self):
        self.section1.add_pageblock_from_dict({
            'block_type': 'Test Block',
            'body': 'test body',
        })
        duplicate = Hierarchy.clone(self.h, 'foo', '/foo/')

        descendants = duplicate.get_root().get_descendants()
        self.assertTrue(descendants[0].label, 'Section 1')
        self.assertTrue(descendants[0].depth, 1)
        self.assertEquals(descendants[0].pageblock_set.count(), 1)

        self.assertTrue(descendants[1].label, 'Section 2')
        self.assertTrue(descendants[1].depth, 1)

        self.assertTrue(descendants[2].label, 'Section 3')
        self.assertTrue(descendants[2].depth, 1)

        self.assertTrue(descendants[3].label, 'Section 4')
        self.assertTrue(descendants[3].depth, 2)

        self.assertTrue(descendants[4].label, 'Section 5')
        self.assertTrue(descendants[4].depth, 3)


class SectionTest(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.section = RootSectionFactory()

    def test_add_pageblock_from_dict(self):
        self.section.add_pageblock_from_dict({
            'block_type': 'Test Block',
            'body': 'test body',
        })
        block = self.section.pageblock_set.first()
        assert(block is not None)
        self.assertEqual(block.block().body, 'test body')
        self.assertEqual(
            self.section.pageblock_set.first(), block,
            'The PageBlock has been added to the Section')

    def test_empty_section_is_unlocked(self):
        self.assertTrue(self.section.unlocked(self.user))

    def test_section_with_test_block_is_unlocked(self):
        self.section.add_pageblock_from_dict({'block_type': 'Test Block'})
        self.assertEqual(self.section.pageblock_set.count(), 1)

        # TestBlock doesn't have an "unlocked" method defined, so this
        # tests that Section.unlocked() handles that case.
        self.assertTrue(self.section.unlocked(self.user))


class TestBlockTest(TestCase):
    def setUp(self):
        self.b = TestBlockFactory()

    def test_is_valid_from_factory(self):
        self.b.full_clean()

    def test_create_from_dict(self):
        testblock = TestBlock.create_from_dict({
            'body': 'abc',
        })
        self.assertEqual(testblock.display_name, 'Test Block')
        self.assertEqual(testblock.body, 'abc')


class UserPageVisitTest(TestCase):
    def test_is_valid_from_factory(self):
        upv = UserPageVisitFactory()
        upv.full_clean()



from __future__ import unicode_literals

from django import template
from django.utils.encoding import smart_text

register = template.Library()


class GetUserSectionStatus(template.Node):
    def __init__(self, user, section, var_name=None):
        self.user = template.Variable(user)
        self.section = template.Variable(section)
        self.var_name = var_name

    def render(self, context):
        u = self.user.resolve(context)
        s = self.section.resolve(context)
        upv = s.get_uservisit(u)
        if self.var_name:
            context[self.var_name] = upv
            return ''
        else:
            return upv


@register.tag('get_user_section_status')
def get_user_section_status(parser, token):
    user = token.split_contents()[1:][0]
    section = token.split_contents()[1:][1]
    var_name = None
    if len(token.split_contents()[1:]) > 2:
        # handle "as some_var"
        var_name = token.split_contents()[1:][3]
    return GetUserSectionStatus(user, section, var_name)


@register.simple_tag
def is_section_unlocked(section, user):
    if hasattr(section, 'unlocked'):
        is_unlocked = section.unlocked(user)
    else:
        is_unlocked = True

    # Convert to 1 or 0 to parse it easily in the template.
    return smart_text(int(is_unlocked))



from __future__ import unicode_literals

"""
Generic View functions.

The idea here is to provide basic functionality. These view functions
should not be routed to directly. Instead, your application should
have its own views which import these functions, enforce their own
authentication/authorization rules, and then call these functions,
overriding behavior where needed.

Ie, you would do something like:

from pagetree.generic.views import PageView, EditView


class MyPageView(PageView):
    def get_extra_context(self):
        ctx = super(MyPageView, self).get_extra_context()
        # Add extra context data
        return ctx

    def dispatch(self, request, *args, **kwargs):
        # Example of a redirect based on user state
        if not request.user.profile.avatar:
            return redirect(reverse('avatar-selector'))

        return super(MyPageView, self).dispatch(request, *args, **kwargs)


class MyEditView(EditView):
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        # Do any additional auth here
        return super(MyEditView, self).dispatch(
            request, *args, **kwargs)

"""
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.shortcuts import render, get_object_or_404
from django.views.generic.base import View, TemplateView
from django.views.generic.edit import FormView

from pagetree.forms import CloneHierarchyForm
from pagetree.helpers import get_section_from_path
from pagetree.models import Hierarchy


def has_responses(section):
    for p in section.pageblock_set.all():
        block = p.block()
        if hasattr(block, 'needs_submit') and block.needs_submit():
            return True

    return False


def visit_root(section, fallback_url="/admin/"):
    """ if they try to visit the root, we need to send them
    either to the first section on the site, or to
    the admin page if there are no sections (so they
    can add some)"""
    if section.get_next():
        # just send them to the first child
        return HttpResponseRedirect(section.get_next().get_absolute_url())
    # no sections available so
    # send them to the fallback
    return HttpResponseRedirect(fallback_url)


def page_submit(section, request):
    proceed = section.submit(request.POST, request.user)
    if proceed:
        next_section = section.get_next()
        if next_section:
            return HttpResponseRedirect(next_section.get_absolute_url())
        else:
            # they are on the "last" section of the site
            # all we can really do is send them back to this page
            return HttpResponseRedirect(section.get_absolute_url())
    # giving them feedback before they proceed
    return HttpResponseRedirect(section.get_absolute_url())


def reset_page(section, request):
    section.reset(request.user)
    return HttpResponseRedirect(section.get_absolute_url())


class UserPageVisitor(object):
    """ simplify/centralize logging visits

    automatically does nothing for an anonymous user
    """
    def __init__(self, section, user):
        self.section = section
        self.user = user

    def visit(self, status=None):
        if self.user.is_anonymous:
            return
        if not status:
            prev = self.section.get_uservisit(self.user)
            if prev is None:
                status = "complete"
                if self.section.needs_submit():
                    status = "incomplete"
            else:
                status = prev.status
        self.section.user_pagevisit(self.user, status)


def generic_view_page(request, path, hierarchy="main",
                      gated=False,
                      template="pagetree/page.html",
                      extra_context=None,
                      no_root_fallback_url="/admin/"
                      ):
    """ generic pagetree page view

    needs the request and path
    hierarchy -> defaults to "main"
    gated -> whether to block users from skipping ahead (ie,
             force sequential access)
    template -> defaults to "pagetree/page.html" so you can either
                override that, or pass in a different one here (or
                just use the basic one that pagetree includes)
    extra_context -> dict of additional variables to pass along to the
                     template
    no_root_fallback -> where to send the user if there are no sections
                        at all in the tree (not even a root). defaults to
                        "/admin/"
    """
    section = get_section_from_path(path, hierarchy=hierarchy)
    root = section.hierarchy.get_root()
    module = section.get_module()
    if section.is_root():
        return visit_root(section, no_root_fallback_url)

    if gated:
        # we need to check that they have visited all previous pages
        # first
        allow, first = section.gate_check(request.user)
        if not allow:
            # redirect to the first one that they need to visit
            return HttpResponseRedirect(first.get_absolute_url())
    upv = UserPageVisitor(section, request.user)
    if request.method == "POST":
        # user has submitted a form. deal with it
        if request.POST.get('action', '') == 'reset':
            upv.visit(status="incomplete")
            return reset_page(section, request)
        upv.visit(status="complete")
        return page_submit(section, request)
    else:
        allow_redo = False
        needs_submit = section.needs_submit()
        if needs_submit:
            allow_redo = section.allow_redo()
        upv.visit()
        instructor_link = has_responses(section)
        context = dict(
            section=section,
            module=module,
            needs_submit=needs_submit,
            allow_redo=allow_redo,
            is_submitted=section.submitted(request.user),
            modules=root.get_children(),
            root=section.hierarchy.get_root(),
            instructor_link=instructor_link,
        )
        if extra_context:
            context.update(extra_context)
        return render(request, template, context)


class SectionMixin(object):
    def get_section(self, path):
        return get_section_from_path(
            path,
            hierarchy_name=self.hierarchy_name,
            hierarchy_base=self.hierarchy_base)

    def get_extra_context(self):
        return self.extra_context

    def perform_checks(self, request, path):
        return None


class PageView(SectionMixin, View):
    template_name = "pagetree/page.html"
    hierarchy_name = "main"
    hierarchy_base = "/"
    extra_context = dict()
    gated = False
    no_root_fallback_url = "/admin/"

    def get_gated(self):
        return self.gated

    def gate_check(self, user):
        if not self.get_gated():
            return None

        # If this view is gated, and we have no user or an anonymous
        # user, then just deny access.
        if (not user) or user.is_anonymous:
            raise PermissionDenied()

        # we need to check that they have visited all previous pages
        # first
        allow, first = self.section.gate_check(user)
        if not allow:
            # redirect to the first one that they need to visit
            return HttpResponseRedirect(first.get_absolute_url())

    def perform_checks(self, request, path):
        self.section = self.get_section(path)
        self.root = self.section.hierarchy.get_root()
        self.module = self.section.get_module()
        if self.section.is_root():
            return visit_root(self.section, self.no_root_fallback_url)
        r = self.gate_check(request.user)
        if r is not None:
            return r
        self.upv = UserPageVisitor(self.section, request.user)
        return None

    def dispatch(self, request, *args, **kwargs):
        path = kwargs['path']
        rv = self.perform_checks(request, path)
        if rv is not None:
            return rv
        return super(PageView, self).dispatch(request, *args, **kwargs)

    def post(self, request, path):
        # user has submitted a form. deal with it
        if request.POST.get('action', '') == 'reset':
            self.upv.visit(status="incomplete")
            return reset_page(self.section, request)
        self.upv.visit(status="complete")
        return page_submit(self.section, request)

    def get(self, request, path):
        allow_redo = False
        needs_submit = self.section.needs_submit()
        if needs_submit:
            allow_redo = self.section.allow_redo()
        self.upv.visit()
        instructor_link = has_responses(self.section)
        context = dict(
            section=self.section,
            module=self.module,
            needs_submit=needs_submit,
            allow_redo=allow_redo,
            is_submitted=self.section.submitted(request.user),
            modules=self.root.get_children(),
            root=self.section.hierarchy.get_root(),
            instructor_link=instructor_link,
        )
        context.update(self.get_extra_context())
        return render(request, self.template_name, context)


def generic_instructor_page(request, path, hierarchy="main",
                            template="pagetree/instructor_page.html",
                            extra_context=None,
                            ):
    """ generic pagetree instructor view

    needs the request and path
    hierarchy -> defaults to "main"
    template -> defaults to "pagetree/instructor_page.html" so you can either
                override that, or pass in a different one here (or
                just use the basic one that pagetree includes)
    extra_context -> dict of additional variables to pass along to the
                     template
    """
    section = get_section_from_path(path, hierarchy=hierarchy)
    root = section.hierarchy.get_root()

    quizzes = []
    for p in section.pageblock_set.all():
        block = p.block()
        if hasattr(block, 'needs_submit') and block.needs_submit():
            return quizzes.append(block)

    context = dict(section=section,
                   quizzes=quizzes,
                   module=section.get_module(),
                   modules=root.get_children(),
                   root=section.hierarchy.get_root())
    if extra_context:
        context.update(extra_context)
    return render(request, template, context)


class InstructorView(SectionMixin, TemplateView):
    template_name = "pagetree/instructor_page.html"
    hierarchy_name = "main"
    hierarchy_base = "/"
    extra_context = dict()

    def dispatch(self, request, *args, **kwargs):
        path = kwargs['path']
        rv = self.perform_checks(request, path)
        if rv is not None:
            return rv
        return super(InstructorView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        path = kwargs['path']
        section = self.get_section(path)
        root = section.hierarchy.get_root()

        quizzes = []
        for p in section.pageblock_set.all():
            block = p.block()
            if hasattr(block, 'needs_submit') and block.needs_submit():
                return quizzes.append(block)

        context = dict(section=section,
                       quizzes=quizzes,
                       module=section.get_module(),
                       modules=root.get_children(),
                       root=section.hierarchy.get_root())
        context.update(self.get_extra_context())
        return context


def generic_edit_page(request, path, hierarchy="main",
                      template="pagetree/edit_page.html",
                      extra_context=None,
                      ):
    """ generic pagetree edit page view

    needs the request and path
    hierarchy -> defaults to "main"
    template -> defaults to "pagetree/edit_page.html" so you can either
                override that, or pass in a different one here (or
                just use the basic one that pagetree includes)
    extra_context -> dict of additional variables to pass along to the
                     template
    """
    section = get_section_from_path(path, hierarchy=hierarchy)
    root = section.hierarchy.get_root()
    context = dict(
        section=section,
        module=section.get_module(),
        modules=root.get_children(),
        available_pageblocks=section.available_pageblocks(),
        root=section.hierarchy.get_root())
    if extra_context:
        context.update(extra_context)
    return render(request, template, context)


class EditView(SectionMixin, TemplateView):
    template_name = "pagetree/edit_page.html"
    hierarchy_name = "main"
    hierarchy_base = "/"
    extra_context = dict()

    def dispatch(self, request, *args, **kwargs):
        path = kwargs['path']
        rv = self.perform_checks(request, path)
        if rv is not None:
            return rv
        return super(EditView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, path):
        section = self.get_section(path)
        root = section.hierarchy.get_root()
        context = dict(
            section=section,
            module=section.get_module(),
            modules=root.get_children(),
            available_pageblocks=section.available_pageblocks(),
            root=section.hierarchy.get_root())
        context.update(self.get_extra_context())
        return context


class CloneHierarchyView(FormView):
    template_name = "pagetree/clone_hierarchy.html"
    form_class = CloneHierarchyForm

    def get_context_data(self, **kwargs):
        context = super(CloneHierarchyView, self).get_context_data(**kwargs)

        hierarchy_id = self.kwargs.get('hierarchy_id')
        context['hierarchy'] = get_object_or_404(Hierarchy, id=hierarchy_id)

        return context

    def form_valid(self, form):
        hierarchy_id = self.kwargs.get('hierarchy_id')
        original = get_object_or_404(Hierarchy, id=hierarchy_id)

        name = form.cleaned_data['name']
        base_url = form.cleaned_data['base_url']

        clone = Hierarchy.clone(original, name, base_url)

        self.success_url = clone.get_root().get_edit_url()
        return super(CloneHierarchyView, self).form_valid(form)



from __future__ import unicode_literals

"""
allow you to pull a pagetree root directly into a template variable.

especially handy for getting navigation/menu stuff into eg, flatpage or
404 templates since you don't have to inject stuff into the views.


Example:

{% load getroot %}

{% getroot main as root %}
{% for section in root.get_children %}
<li>... top level sections ...</li>
{% endfor %}

"""
from django import template
from pagetree.helpers import get_hierarchy

register = template.Library()


class GetRootNode(template.Node):
    def __init__(self, hierarchy, var_name):
        self.hierarchy = hierarchy
        self.var_name = var_name

    def render(self, context):
        h = get_hierarchy(self.hierarchy)
        r = h.get_root()
        context[self.var_name] = r
        return ''


@register.tag('getroot')
def getroot(parser, token):
    hierarchy = token.split_contents()[1:][0]
    var_name = token.split_contents()[1:][2]
    return GetRootNode(hierarchy, var_name)



from __future__ import unicode_literals

import os.path

from django.conf.urls import url
import django.views.static

from pagetree.generic.views import CloneHierarchyView
from .views import (
    reorder_pageblocks, reorder_section_children, add_child_section,
    move_section, revert_to_version, exporter, create_tree_root,
    delete_pageblock, edit_section, delete_section, add_pageblock,
    edit_pageblock, export_pageblock_json, import_pageblock_json,
)


media_root = os.path.join(os.path.dirname(__file__), "media")

urlpatterns = [
    url(r'^media/(?P<path>.*)$', django.views.static.serve,
        {'document_root': media_root}),
]

urlpatterns += [
    url(r'^reorder_pageblocks/(?P<section_id>\d+)/$', reorder_pageblocks, {},
        "reorder-pageblocks"),
    url(r'^reorder_section_children/(?P<section_id>\d+)/$',
        reorder_section_children, {}, "reorder-section-children"),
    url(r'^section/add/(?P<section_id>\d+)/$', add_child_section, {},
        "add-child-section"),
    url(r'^section/move/(?P<section_id>\d+)/$', move_section, {},
        "move-section"),
    url(r'^pageblock/add/(?P<section_id>\d+)/$', add_pageblock, {},
        "add-pageblock"),
    url(r'^pageblock/edit/(?P<pageblock_id>\d+)/$', edit_pageblock, {},
        "edit-pageblock"),
    url(r'^pageblock/jsonexport/(?P<pageblock_id>\d+)/$',
        export_pageblock_json, {}, "export-pageblock-json"),
    url(r'^pageblock/jsonimport/(?P<pageblock_id>\d+)/$',
        import_pageblock_json, {}, "import-pageblock-json"),
    url(r'^delete_section/(?P<section_id>\d+)/$', delete_section, {},
        "delete-section"),
    url(r'^edit_section/(?P<section_id>\d+)/$', edit_section, {},
        "edit-section"),
    url(r'^delete_pageblock/(?P<pageblock_id>\d+)/$', delete_pageblock, {},
        "delete-pageblock"),
    url(r'^create_tree_root$', create_tree_root, {}, "create_tree_root"),
    url(r'^export/$', exporter, {}, 'export-hierarchy'),
    url(r'^version/(?P<version_id>\d+)/revert/$', revert_to_version, {},
        "revert-to-version"),
    url(r'^clone_hierarchy/(?P<hierarchy_id>\d+)/$',
        CloneHierarchyView.as_view(), name="clone-hierarchy"),
]



from __future__ import unicode_literals

from django.test import TestCase
from pagetree.forms import CloneHierarchyForm
from pagetree.tests.factories import HierarchyFactory


class CloneHierarchyFormTest(TestCase):
    def test_init(self):
        h = HierarchyFactory()
        CloneHierarchyForm(instance=h)

    def test_clean(self):
        h = HierarchyFactory()
        f = CloneHierarchyForm({
            'name': 'new name',
            'base_url': 'new',
        }, instance=h)
        self.assertTrue(f.is_valid())
        new_hier = f.save()
        self.assertEqual(new_hier.name, 'new name')
        self.assertEqual(new_hier.base_url, 'new')

    def test_clean_no_base_url(self):
        h = HierarchyFactory()
        f = CloneHierarchyForm({
            'name': 'new name',
            'base_url': '',
        }, instance=h)
        self.assertTrue(f.is_valid())
        new_hier = f.save()
        self.assertEqual(new_hier.name, 'new name')
        self.assertEqual(new_hier.base_url, 'new-name')

    def test_clean_prevents_duplicate_name(self):
        h = HierarchyFactory()
        f = CloneHierarchyForm({
            'name': h.name,
            'base_url': 'new',
        }, instance=h)
        self.assertFalse(f.is_valid())
        self.assertEqual(f.errors, {
            '__all__': [
                u"There's already a hierarchy with the name: main"]
        })

    def test_clean_prevents_duplicate_base_url(self):
        h = HierarchyFactory()
        f = CloneHierarchyForm({
            'name': 'new name',
            'base_url': h.base_url,
        }, instance=h)
        self.assertFalse(f.is_valid())
        self.assertEqual(f.errors, {
            '__all__': [
                u"There's already a hierarchy with the base_url: /pages/"]
        })



from __future__ import unicode_literals

import random
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import models
from django import forms
from django.template.loader import get_template
from django.http import Http404
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.template.defaultfilters import slugify
from json import dumps
from treebeard.mp_tree import MP_Node
import django.core.exceptions
from treebeard.forms import movenodeform_factory


# dummy it out
class dummystatsd(object):
    def incr(self, metric):
        pass


statsd = dummystatsd()


try:
    from django_statsd.clients import statsd
except ImportError:
    pass


class Hierarchy(models.Model):
    name = models.CharField(max_length=256)
    base_url = models.CharField(max_length=256, default="")

    @staticmethod
    def get_hierarchy(name):
        return Hierarchy.objects.get_or_create(name=name,
                                               defaults=dict(base_url="/"))[0]

    def get_absolute_url(self):
        return self.base_url

    def __str__(self):
        return self.name

    def get_root(self):
        # will create it if it doesn't exist
        try:
            return Section.objects.get(hierarchy=self,
                                       label="Root").get_root()
        except Section.DoesNotExist:
            return Section.add_root(label="Root", slug="", hierarchy=self)

    def get_top_level(self):
        return self.get_root().get_children()

    def find_section_from_path(self, path):
        if path.endswith("/"):
            path = path[:-1]
        root = self.get_root()
        current = root
        if path == "":
            return current
        for slug in path.split("/"):
            slugs = dict()
            for c in current.get_children():
                slugs[c.slug] = c
            if slug in slugs:
                current = slugs[slug]
            else:
                return None
        return current

    def get_section_from_path(self, path):
        s = self.find_section_from_path(path)
        if s is None:
            raise Http404()
        return s

    @staticmethod
    def available_pageblocks():
        if hasattr(settings, 'PAGEBLOCKS') and \
           settings.PAGEBLOCKS is not None:
            models = [
                apps.get_model(*pb.split('.'))
                for pb in settings.PAGEBLOCKS]
            return models
        else:
            return []

    def get_first_leaf(self, section):
        if section.is_leaf():
            return section
        return self.get_first_leaf(section.get_children()[0])

    def get_last_leaf(self, section):
        if section.is_leaf():
            return section
        return self.get_last_leaf(list(section.get_children())[-1])

    def as_dict(self):
        return dict(name=self.name,
                    base_url=self.base_url,
                    sections=[self.get_root().as_dict()])

    def add_section_from_dict(self, d):
        s = Section.add_root(label="Root", slug="", hierarchy=self)
        for pb in d.get('pageblocks', []):
            s.add_pageblock_from_dict(pb)
        for c in d.get('children', []):
            s.add_child_section_from_dict(c)
        return s

    @staticmethod
    def from_dict(d):
        h = Hierarchy.objects.create(name=d.get('name', ''),
                                     base_url=d.get('base_url', '/'))
        for s in d.get('sections', []):
            h.add_section_from_dict(s)
        return h

    def get_user_location(self, user):
        if user.is_anonymous:
            return "/"
        (ul, created) = UserLocation.objects.get_or_create(
            user=user,
            hierarchy=self)
        return ul.path

    def get_user_section(self, user):
        location = self.get_user_location(user)[len(self.base_url):]
        return self.find_section_from_path(location)

    def user_visit(self, user, section):
        path = section.get_absolute_url()
        (ul, created) = UserLocation.objects.get_or_create(
            user=user,
            hierarchy=self)
        ul.path = path
        ul.save()

    @staticmethod
    def clone(original, name, base_url):
        hierarchy = Hierarchy.objects.create(name=name, base_url=base_url)
        root = Section.add_root(label="Root", slug="", hierarchy=hierarchy)
        Section.clone(original.get_root(), root)
        return hierarchy


class Section(MP_Node):
    label = models.CharField(max_length=256)
    slug = models.SlugField()
    hierarchy = models.ForeignKey(Hierarchy, on_delete=models.CASCADE)
    show_toc = models.BooleanField(
        default=False,
        help_text=("list table of contents of "
                   "immediate child sections (if applicable)"))
    deep_toc = models.BooleanField(
        default=False,
        help_text=("include children of children in TOC. "
                   "(this only makes sense if the above is checked)"))

    def enforce_slug(self):
        # root node gets an empty slug
        if self.is_root():
            self.slug = ""
        # but no other section should ever be able to
        self.slug = slugify(self.slug)[:50]
        if self.slug == "":
            self.slug = slugify(self.label)[:50]
        if self.slug == "":
            # label is empty too, so we need to fall way back
            self.slug = "section-%d" % self.id
        # enforce uniqueness at this section's level in the
        # tree. ie, don't allow two siblings to have the same slug.
        # duplicate slugs at different levels of the tree are fine.
        if not self.is_root():
            # need to make sure no other siblings have the same slug
            while self.slug in self.get_sibling_slugs():
                self.slug = self.slug + "-%d" % self.id

        # TODO: this should maybe go on a save hook to make sure
        # it is always enforced.
        self.save()

    def clear_caches(self):
        cache.delete("pagetree.%d.get_absolute_url" % self.id)
        cache.delete("pagetree.%d.is_last_child" % self.id)
        cache.delete("pagetree.%d.get_edit_url" % self.id)
        cache.delete("pagetree.%d.get_tree_depth" % self.id)

        if hasattr(settings, 'PAGETREE_CUSTOM_CACHE_CLEAR') and \
           callable(settings.PAGETREE_CUSTOM_CACHE_CLEAR):
            settings.PAGETREE_CUSTOM_CACHE_CLEAR(self)

    def clear_tree_cache(self):
        depth_first_traversal = self.get_root().get_annotated_list()
        for (s, ai) in depth_first_traversal:
            s.clear_caches()
        statsd.incr("pagetree.cache_cleared")

    def get_sibling_slugs(self):
        return [s.slug for s in self.get_siblings() if s != self]

    def get_module(self):
        """ get the top level module that the section is in"""
        if self.is_root():
            return None
        if self.depth == 2:
            return self
        return self.get_ancestors()[1]

    def get_descendants(self):
        return MP_Node.get_descendants(self).select_related('hierarchy')

    def is_first_child(self):
        return self.get_first_sibling().id == self.id

    def is_last_child(self):
        key = "pagetree.%d.is_last_child" % self.id
        v = cache.get(key)
        if v is not None:
            statsd.incr("pagetree.is_last_child.cache_hit")
            return v
        statsd.incr("pagetree.is_last_child.cache_miss")
        v = self._is_last_child()
        cache.set(key, v)
        return v

    def _is_last_child(self):
        return self.get_last_sibling().id == self.id

    def closing_children(self):
        """ this returns the list of adjacent last children.
        we need this to know how many levels deep need to be closed
        when making the menus as flattened nested lists.
        look for 'closing_children' in a menu.html in a project
        that uses pagetree to see exactly what i mean. """

        s = self
        while not s.is_root() and s.is_last_child():
            yield s
            s = s.get_parent()

    def get_previous(self):
        # previous node in the depth-first traversal
        depth_first_traversal = self.get_root().get_annotated_list()
        for (i, (s, ai)) in enumerate(depth_first_traversal):
            if s.id == self.id:
                # make sure we don't return the root
                if i > 1 and not depth_first_traversal[i - 1][0].is_root():
                    return depth_first_traversal[i - 1][0]
                else:
                    return None
        # made it through without finding ourselves? weird.
        return None

    def get_next(self):
        # next node in the depth-first traversal
        depth_first_traversal = self.get_root().get_annotated_list()
        for (i, (s, ai)) in enumerate(depth_first_traversal):
            if s.id == self.id:
                if i < len(depth_first_traversal) - 1:
                    return depth_first_traversal[i + 1][0]
                else:
                    return None
        # made it through without finding ourselves? weird.
        return None

    def append_child(self, label, slug='', show_toc=False, deep_toc=False):
        if slug == '':
            slug = slugify(label)
        self.save()
        c = self.add_child(label=label, slug=slug, hierarchy=self.hierarchy,
                           show_toc=show_toc, deep_toc=deep_toc)
        c.enforce_slug()
        return c

    def append_pageblock(self, label, css_extra, content_object):
        neword = self.pageblock_set.count() + 1
        return PageBlock.objects.create(section=self, label=label,
                                        ordinality=neword,
                                        css_extra=css_extra,
                                        content_object=content_object)

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        key = "pagetree.%d.get_absolute_url" % self.id
        v = cache.get(key)
        if v is not None:
            statsd.incr("pagetree.get_absolute_url.cache_hit")
            return v
        statsd.incr("pagetree.get_absolute_url.cache_miss")
        v = self._get_absolute_url()
        cache.set(key, v)
        return v

    def _get_absolute_url(self):
        ancestors = self.get_ancestors()
        slugs = [a.slug for a in ancestors[1:]]
        if len(slugs) == 0:
            return self.hierarchy.get_absolute_url() + self.slug + "/"
        url = (self.hierarchy.get_absolute_url() + "/".join(slugs) +
               "/" + self.slug + "/")
        return url

    def get_edit_url(self):
        key = "pagetree.%d.get_edit_url" % self.id
        v = cache.get(key)
        if v is not None:
            statsd.incr("pagetree.get_edit_url.cache_hit")
            return v
        statsd.incr("pagetree.get_edit_url.cache_miss")
        v = self._get_edit_url()
        cache.set(key, v)
        return v

    def _get_edit_url(self):
        ancestors = self.get_ancestors()
        slugs = [a.slug for a in ancestors[1:]]
        if len(slugs) == 0:
            return (
                self.hierarchy.get_absolute_url() + "edit/" + self.slug + "/")
        url = (self.hierarchy.get_absolute_url() + "edit/" + "/".join(slugs) +
               "/" + self.slug + "/")
        return url

    def get_instructor_url(self):
        if self.is_root():
            return self.hierarchy.get_absolute_url() + "instructor/"
        return self.get_parent().get_instructor_url() + self.slug + "/"

    def get_path(self):
        """ same as get_absolute_url, without the leading /"""
        return self.get_absolute_url()[len(self.hierarchy.base_url):]

    def add_child_section_form(self):
        class AddChildSectionForm(forms.Form):
            label = forms.CharField()

        return AddChildSectionForm()

    def renumber_pageblocks(self):
        i = 1
        for block in self.pageblock_set.all():
            block.ordinality = i
            block.save()
            i += 1

    def edit_form(self):
        class EditSectionForm(forms.Form):
            label = forms.CharField(initial=self.label)
            slug = forms.CharField(initial=self.slug)
            show_toc = forms.BooleanField(
                initial=self.show_toc,
                required=False,
                label="Show Table of Contents",
                help_text=("list table of contents of "
                           "immediate child sections (if applicable)"))
            deep_toc = forms.BooleanField(
                initial=self.deep_toc,
                required=False,
                label="Show Deep Table of Contents",
                help_text=(
                    "include children of children (etc) in TOC. "
                    "(this only makes sense if the above is enabled)"))
        return EditSectionForm()

    def move_form(self):
        MoveSectionForm = movenodeform_factory(
            type(self),
            exclude=('label', 'slug', 'hierarchy', 'show_toc', 'deep_toc'))
        return MoveSectionForm(instance=self)

    def update_children_order(self, children_ids):
        """children_ids is a list of Section ids for the children
        in the order that they should be set to.

        use with caution. if the ids in children_ids don't match up
        right it will break or do strange things.
        """
        for section_id in children_ids:
            s = Section.objects.get(id=section_id)
            p = s.get_parent()
            s.move(p, pos="last-child")
        self.clear_tree_cache()
        return

    def update_pageblocks_order(self, pageblock_ids):
        """pageblock_ids is a list of PageBlock ids for the children
        in the order that they should be set to.

        use with caution. if the ids in pageblock_ids don't match up
        right it will break or do strange things.
        """
        for (i, id) in enumerate(pageblock_ids):
            sc = PageBlock.objects.get(id=id)
            sc.ordinality = i + 1
            sc.save()

    def available_pageblocks(self):
        return self.hierarchy.available_pageblocks()

    def add_pageblock_form(self):
        # This unique id should instead be derived from the block type's
        # ID, but that requires a new templatetag. This works for now.
        unique_id = random.randint(0, 10000)

        class AddPageBlockForm(forms.Form):
            label = forms.CharField(
                required=False,
                widget=forms.TextInput(
                    attrs={'id': 'id_label_%d' % unique_id}
                ))
            css_extra = forms.CharField(
                label='extra CSS classes',
                required=False,
                widget=forms.TextInput(
                    attrs={'id': 'id_css_extra_%d' % unique_id}
                ))

        return AddPageBlockForm()

    def get_first_leaf(self):
        if self.is_leaf():
            return self
        return self.get_children()[0].get_first_leaf()

    def get_last_leaf(self):
        if self.is_leaf():
            return self
        return list(self.get_children())[-1].get_last_leaf()

    def reset(self, user):
        """ clear a user's responses to all pageblocks on this page """
        for p in self.pageblock_set.all():
            block = p.block()
            if hasattr(block, 'needs_submit'):
                if block.needs_submit():
                    block.clear_user_submissions(user)

    def submit(self, request_data, user):
        """ store users's responses to the pageblocks on this page """
        proceed = True
        for p in self.pageblock_set.all():
            block = p.block()
            if hasattr(block, 'needs_submit'):
                if block.needs_submit():
                    prefix = "pageblock-%d-" % p.id
                    data = dict()
                    for k in request_data.keys():
                        if k.startswith(prefix):
                            # handle lists for multi-selects
                            v = request_data.getlist(k)
                            if len(v) == 1:
                                data[k[len(prefix):]] = request_data[k]
                            else:
                                data[k[len(prefix):]] = v
                    block.submit(user, data)
                    if hasattr(block, 'redirect_to_self_on_submit'):
                        # semi bug here?
                        # proceed will only be set by the last submittable
                        # block on the page. previous ones get ignored.
                        proceed = not block.redirect_to_self_on_submit()
        return proceed

    def needs_submit(self):
        """ if any blocks on the page need to be submitted """
        for p in self.pageblock_set.all():
            block = p.block()
            if hasattr(block, 'needs_submit'):
                if block.needs_submit():
                    return True
        return False

    def allow_redo(self, **kwargs):
        """ if any blocks on the page are allowed to be resubmitted """
        for p in self.pageblock_set.all():
            block = p.block()
            if hasattr(block, 'allow_redo'):
                if hasattr(block.allow_redo, '__call__'):
                    if block.allow_redo(**kwargs):
                        return True
                else:
                    # not callable, so we expect it
                    # to just be a boolean property
                    if block.allow_redo:
                        return True
        return False

    def submitted(self, user):
        """ if all blocks on the page that require submissions
        have been submitted """
        for p in self.pageblock_set.all():
            block = p.block()
            if hasattr(block, 'needs_submit'):
                if block.needs_submit():
                    try:
                        s = block.unlocked(user)
                        if not s:
                            # there's an unsubmitted block
                            return False
                    except (AttributeError, ValueError):
                        # most likely: no unlocked() method
                        pass
        # made it all the way through without any blocks
        # reporting that they are unsubmitted
        return True

    def unlocked(self, user):
        """Returns True if all the blocks on this page are unlocked."""

        for p in self.pageblock_set.all():
            block = p.block()
            if hasattr(block, 'unlocked'):
                if not block.unlocked(user):
                    return False

        return True

    def as_dict(self):
        return dict(
            label=self.label,
            slug=self.slug,
            show_toc=self.show_toc,
            deep_toc=self.deep_toc,
            pageblocks=[b.as_dict() for b in self.pageblock_set.all()],
            children=[s.as_dict() for s in self.get_children()],
        )

    def from_dict(self, d):
        self.label = d.get('label', '')
        self.slug = d.get('slug', '')
        self.show_toc = d.get('show_toc', False)
        self.deep_toc = d.get('deep_toc', False)
        self.save()
        for pb in d.get('pageblocks', []):
            self.add_pageblock_from_dict(pb)
        for c in d.get('children', []):
            self.add_child_section_from_dict(c)

    def add_pageblock_from_dict(self, d):
        target_type = d.get('block_type', '')

        # now we need to figure out which kind of pageblock to create
        found_pbclass = None
        for pb_class in self.available_pageblocks():
            if not pb_class:
                continue
            if pb_class.display_name == target_type:
                # a match
                found_pbclass = pb_class
                break

        if found_pbclass:
            if hasattr(found_pbclass, 'create_from_dict'):
                # Prepare a dictionary for the custom pageblock's
                # create_from_dict method by removing the data that's
                # only used elsewhere.
                block_dict = d.copy()
                block_dict.pop('block_type', None)
                block_dict.pop('label', None)
                block_dict.pop('css_extra', None)
                block = found_pbclass.create_from_dict(block_dict)

                self.append_pageblock(label=d.get('label', ''),
                                      css_extra=d.get('css_extra', ''),
                                      content_object=block)

    def add_child_section_from_dict(self, d):
        s = self.append_child(
            d.get('label', ''), d.get('slug', ''),
            d.get('show_toc', False), d.get('deep_toc', False))
        s.save()
        for pb in d.get('pageblocks', []):
            s.add_pageblock_from_dict(pb)
        for c in d.get('children', []):
            s.add_child_section_from_dict(c)
        self.clear_tree_cache()

    def user_visit(self, user):
        self.hierarchy.user_visit(user, self)

    def user_pagevisit(self, user, status="incomplete"):
        try:
            (upv, created) = UserPageVisit.objects.get_or_create(
                section=self,
                user=user)
        except django.core.exceptions.MultipleObjectsReturned:
            upv = UserPageVisit.objects.filter(
                section=self, user=user).first()
        upv.status = status
        upv.save()

    def get_uservisit(self, user):
        # returns None if the userpagevisit doesn't exist
        return self.userpagevisit_set.filter(user=user).first()

    def save_version(self, user, activity=None, comment=None):
        if not activity:
            activity = ""
        if not comment:
            comment = ""
        json = dumps(self.as_dict())
        v = Version.objects.create(
            section=self,
            user=user,
            data=json,
            activity=activity,
            comment=comment)
        self.clear_tree_cache()
        return v

    def gate_check(self, user):
        """ return bool, section tuple for whether the user
        has visited every section previous to this one and
        which is the first that they need to visit if that's
        not the case """
        if not user:
            # no user: the important thing is just that we deny access
            return False, self

        # otherwise, let's start at the beginning and check each
        depth_first_traversal = self.get_annotated_list(parent=self.get_root())

        # prep a list of all the visits for this user
        upvs = [upv.section.id
                for upv in list(UserPageVisit.objects.filter(user=user))
                if upv.status == 'complete']
        for (i, (s, ai)) in enumerate(depth_first_traversal):
            # skip the root
            if s.is_root():
                continue
            if s.id == self.id:
                # we've reached the current section. That
                # means they're good to go.
                return True, None
            else:
                if s.id not in upvs:
                    # uh oh. found a page that they haven't visited
                    # need to send them there first
                    return False, s
        # went through the entire list of sections without finding
        # the current section?!
        assert False, "current section not found in traversal"

    @staticmethod
    def clone(original, section):
        for b in original.pageblock_set.all():
            section.add_pageblock_from_dict(b.as_dict())
        section.save()

        children = original.get_children()
        for old_child in children:
            child = section.append_child(label=old_child.label,
                                         slug=old_child.slug,
                                         show_toc=old_child.show_toc,
                                         deep_toc=old_child.deep_toc)

            Section.clone(old_child, child)

        return section

    def get_tree_depth(self):
        key = 'pagetree.{}.get_tree_depth'.format(self.pk)
        v = cache.get(key)
        if v is not None:
            return v

        for idx, sec in enumerate(self.get_tree()):
            if sec == self:
                cache.set(key, idx)
                return idx


class PageBlock(models.Model):
    section = models.ForeignKey(Section, on_delete=models.CASCADE)
    ordinality = models.PositiveIntegerField(default=1)
    label = models.CharField(max_length=256, blank=True, null=True)
    css_extra = models.CharField(
        max_length=256, blank=True, null=True,
        help_text="extra CSS classes (space separated)")

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')

    class Meta:
        ordering = ('section', 'ordinality',)

    def __str__(self):
        return "%s [%d]: %s" % (self.section.label, self.ordinality,
                                self.label)

    def block(self):
        return self.content_object

    def edit_label(self):
        """ provide a label for the pageblock to make the
        edit interface easier to read """
        block = self.block()
        if hasattr(block, 'edit_label'):
            return block.edit_label()
        else:
            return block.display_name

    def render(self, **kwargs):
        if hasattr(self.content_object, "template_file"):
            t = get_template(getattr(self.content_object, "template_file"))
            d = kwargs
            d['pageblock'] = self.content_object
            d['block'] = d['pageblock']
            return t.render(d)
        else:
            return self.content_object.render()

    def render_js(self, **kwargs):
        if hasattr(self.content_object, "js_template_file"):
            t = get_template(getattr(self.content_object, "js_template_file"))
            d = kwargs
            d['block'] = self.content_object
            return t.render(d)
        elif hasattr(self.content_object, "js_render"):
            return self.content_object.js_render()
        else:
            return ""

    def render_css(self, **kwargs):
        if hasattr(self.content_object, "css_template_file"):
            t = get_template(getattr(self.content_object, "css_template_file"))
            d = kwargs
            d['block'] = self.content_object
            return t.render(d)
        elif hasattr(self.content_object, "css_render"):
            return self.content_object.css_render()
        else:
            return ""

    def render_summary(self, **kwargs):
        if hasattr(self.content_object, "summary_template_file"):
            t = get_template(getattr(self.content_object,
                                     "summary_template_file"))
            d = kwargs
            d['block'] = self.content_object
            return t.render(d)
        elif hasattr(self.content_object, "summary_render"):
            return self.content_object.summary_render()
        else:
            return ""

    def default_edit_form(self):
        class EditForm(forms.Form):
            label = forms.CharField(
                initial=self.label,
                required=False,
                widget=forms.TextInput(
                    attrs={'id': 'id_label_%d' % self.pk}
                ))
            css_extra = forms.CharField(
                initial=self.css_extra,
                label='extra CSS classes',
                required=False,
                widget=forms.TextInput(
                    attrs={'id': 'id_css_extra_%d' % self.pk}
                ))
        return EditForm()

    def edit_form(self):
        return self.content_object.edit_form()

    def edit(self, vals, files):
        self.label = vals.get('label', '')
        self.css_extra = vals.get('css_extra', '')
        self.save()
        self.content_object.edit(vals, files)

    def delete(self):
        section = self.section
        super(PageBlock, self).delete()  # Call the "real" delete() method
        section.renumber_pageblocks()

    def as_dict(self):
        d = dict()
        if hasattr(self.content_object, 'as_dict'):
            d = self.content_object.as_dict()
        d['label'] = self.label
        d['css_extra'] = self.css_extra
        d['block_type'] = self.content_object.display_name
        return d

    def import_from_dict(self, d):
        self.label = d.get('label', '')
        self.css_extra = d.get('css_extra', '')
        self.save()

        if hasattr(self.content_object, 'import_from_dict'):
            self.content_object.import_from_dict(d)

    @classmethod
    def create_from_dict(cls, d):
        return cls.objects.create(**d)


class UserLocation(models.Model):
    """ last path a given user visited (for a particular hierarchy) """
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    hierarchy = models.ForeignKey(Hierarchy, on_delete=models.CASCADE)
    path = models.CharField(max_length=256, default="/")

    class Meta:
        unique_together = (('user', 'hierarchy'),)


class UserPageVisit(models.Model):
    """ for detailed tracking """
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    section = models.ForeignKey(Section, on_delete=models.CASCADE)
    status = models.CharField(max_length=256, default="incomplete")
    first_visit = models.DateTimeField(auto_now_add=True)
    last_visit = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('user', 'section'),)


class Version(models.Model):
    """ very basic versioning """
    section = models.ForeignKey(Section, on_delete=models.CASCADE)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    saved_at = models.DateTimeField(auto_now_add=True)
    activity = models.TextField(default="", blank=True, null=True)
    data = models.TextField(default="", blank=True, null=True)
    comment = models.TextField(default="", blank=True, null=True)

    class Meta:
        ordering = ["-saved_at", ]

    def __lt__(self, other):
        return self.saved_at < other.saved_at

    def more_recent_versions(self):
        versions = list(
            self.section.version_set.filter(
                saved_at__gt=self.saved_at))
        for s in self.section.get_descendants():
            for v in s.version_set.filter(saved_at__gt=self.saved_at):
                versions.append(v)
        versions.sort()
        return versions



from __future__ import unicode_literals

from django.test import TestCase
from django.test.client import RequestFactory
from django.test.client import Client
from django.http import Http404
from django.contrib.auth.models import User
from pagetree.models import Hierarchy
from pagetree.generic.views import has_responses, visit_root
from pagetree.generic.views import UserPageVisitor
from pagetree.generic.views import generic_view_page
from pagetree.generic.views import generic_instructor_page
from pagetree.generic.views import generic_edit_page


class Anon(object):
    is_anonymous = True


class NonAnon(object):
    is_anonymous = False


class MockSection(object):
    def __init__(self, needs_submit_value=False):
        self.needs_submit_called = 0
        self.user_pagevisit_called = 0
        self.needs_submit_value = needs_submit_value

    def needs_submit(self):
        self.needs_submit_called += 1
        return self.needs_submit_value

    def user_pagevisit(self, user, status):
        self.user_pagevisit_called += 1
        self.user = user
        self.status = status

    def get_uservisit(self, user):
        return None


class UserPageVisitorTest(TestCase):
    def test_visit_anonymous(self):
        s = MockSection()
        upv = UserPageVisitor(s, Anon())
        upv.visit()
        self.assertEqual(s.needs_submit_called, 0)

    def test_visit_not_status(self):
        s = MockSection(needs_submit_value=False)
        upv = UserPageVisitor(s, NonAnon())
        upv.visit(status=False)
        self.assertEqual(s.needs_submit_called, 1)
        self.assertEqual(s.user_pagevisit_called, 1)
        self.assertEqual(s.status, "complete")

    def test_visit_not_status_needs_submit(self):
        s = MockSection(needs_submit_value=True)
        upv = UserPageVisitor(s, NonAnon())
        upv.visit(status=False)
        self.assertEqual(s.needs_submit_called, 1)
        self.assertEqual(s.user_pagevisit_called, 1)
        self.assertEqual(s.status, "incomplete")

    def test_visit_status(self):
        s = MockSection()
        upv = UserPageVisitor(s, NonAnon())
        upv.visit(status=True)
        self.assertEqual(s.needs_submit_called, 0)
        self.assertEqual(s.user_pagevisit_called, 1)
        self.assertEqual(s.status, True)


class OneLevelDeepTest(TestCase):
    def setUp(self):
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'section-2',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 3',
                'slug': 'section-3',
                'pageblocks': [],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.section2 = r[1]
        self.section3 = r[2]

    def tearDown(self):
        self.h.delete()

    def test_has_responses(self):
        self.assertFalse(has_responses(self.section1))

    def test_visit_root(self):
        r = visit_root(self.section1)
        self.assertEqual(r.status_code, 302)

    def test_visit_root_fallback(self):
        r = visit_root(self.section3)
        self.assertEqual(r.status_code, 302)


class GenericPageViewTest(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'section-2',
                'pageblocks': [
                    {'label': 'Welcome to your new Forest Site',
                     'css_extra': '',
                     'block_type': 'Test Block',
                     'body': 'You should now use the edit link to add content',
                     },
                ],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 3',
                'slug': 'section-3',
                'pageblocks': [],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.section2 = r[1]
        self.section3 = r[2]
        self.u = User.objects.create(username="test")
        self.u.set_password("test")
        self.u.save()

    def test_nonexistant(self):
        request = self.factory.get("/foo/")
        with self.assertRaises(Http404):
            generic_view_page(request, "/foo/")

    def test_root(self):
        request = self.factory.get("/")
        response = generic_view_page(request, "/")
        self.assertEqual(response.status_code, 302)

    def test_section1(self):
        request = self.factory.get("/section-1/")
        request.user = self.u
        response = generic_view_page(
            request, "section-1/",
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 200)

    def test_section1_gated(self):
        request = self.factory.get("/section-1/")
        request.user = self.u
        response = generic_view_page(
            request, "section-1/",
            gated=True,
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 200)

    def test_section2(self):
        request = self.factory.get("/section-2/")
        request.user = self.u
        response = generic_view_page(
            request, "section-2/",
            extra_context=dict(foo='bar'),
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 200)

    def test_section2_gated(self):
        request = self.factory.get("/section-2/")
        request.user = self.u
        response = generic_view_page(
            request, "section-2/",
            gated=True,
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 302)

    def test_section1_post(self):
        request = self.factory.post("/section-1/", dict())
        request.user = self.u
        response = generic_view_page(
            request, "section-1/",
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 302)

    def test_section1_reset(self):
        request = self.factory.post("/section-1/", dict(action="reset"))
        request.user = self.u
        response = generic_view_page(
            request, "section-1/",
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 302)

    def test_section3_post(self):
        request = self.factory.post("/section-3/", dict())
        request.user = self.u
        response = generic_view_page(
            request, "section-3/",
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 302)

    def test_instructor_section1(self):
        request = self.factory.get("/instructor/section-1/")
        request.user = self.u
        response = generic_instructor_page(
            request, "section-1/",
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 200)

    def test_edit_section1(self):
        request = self.factory.get("/edit/section-1/")
        request.user = self.u
        response = generic_edit_page(
            request, "section-1/",
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 200)

    def test_instructor_section1_extra(self):
        request = self.factory.get("/instructor/section-1/")
        request.user = self.u
        response = generic_instructor_page(
            request, "section-1/",
            extra_context=dict(foo="bar"),
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 200)

    def test_edit_section1_extra(self):
        request = self.factory.get("/edit/section-1/")
        request.user = self.u
        response = generic_edit_page(
            request, "section-1/",
            extra_context=dict(foo="bar"),
            template="pagetree/test_page.html")
        self.assertEqual(response.status_code, 200)


class GenericPageViewURLConfTest(TestCase):
    def setUp(self):
        self.h = Hierarchy.objects.create(name="main", base_url="")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'section-1',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'section-2',
                'pageblocks': [
                    {'label': 'Welcome to your new Forest Site',
                     'css_extra': '',
                     'block_type': 'Test Block',
                     'body': 'You should now use the edit link to add content',
                     },
                ],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 3',
                'slug': 'section-3',
                'pageblocks': [],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.section2 = r[1]
        self.section3 = r[2]
        self.u = User.objects.create(username="test")
        self.u.set_password("test")
        self.u.save()
        self.c = Client()
        self.c.login(username="test", password="test")

    def test_edit_section1(self):
        response = self.c.get("/pages/edit/section-1/")
        self.assertEqual(response.status_code, 200)

    def test_instructor_section1(self):
        response = self.c.get("/pages/instructor/section-1/")
        self.assertEqual(response.status_code, 200)

    def test_section1(self):
        response = self.c.get("/pages/section-1/")
        self.assertEqual(response.status_code, 200)


class GenericPageViewURLConfTestSecond(TestCase):
    """ make sure non default hierarchies also work """
    def setUp(self):
        self.h = Hierarchy.objects.create(name="two", base_url="/pages2")
        self.root = self.h.get_root()
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 1',
                'slug': 'sec-1',
                'pageblocks': [],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 2',
                'slug': 'sec-2',
                'pageblocks': [
                    {'label': 'Welcome to your new Forest Site',
                     'css_extra': '',
                     'block_type': 'Test Block',
                     'body': 'You should now use the edit link to add content',
                     },
                ],
                'children': [],
            })
        self.root.add_child_section_from_dict(
            {
                'label': 'Section 3',
                'slug': 'sec-3',
                'pageblocks': [],
                'children': [],
            })
        r = self.root.get_children()
        self.section1 = r[0]
        self.section2 = r[1]
        self.section3 = r[2]
        self.u = User.objects.create(username="test")
        self.u.set_password("test")
        self.u.save()
        self.c = Client()
        self.c.login(username="test", password="test")

    def test_edit_section1(self):
        response = self.c.get("/pages2/edit/sec-1/")
        self.assertEqual(response.status_code, 200)

    def test_instructor_section1(self):
        response = self.c.get("/pages2/instructor/sec-1/")
        self.assertEqual(response.status_code, 200)

    def test_section1(self):
        response = self.c.get("/pages2/sec-1/")
        self.assertEqual(response.status_code, 200)



from __future__ import unicode_literals

""" render templatetag

let's us do {% render block %}

instead of {{block.render}}

and have block.render() get called with the existing
request context passed through

"""

from six import string_types
from django import template

register = template.Library()


class BaseNode(template.Node):
    def __init__(self, block):
        self.block = block

    def prepare_context_data(self, context):
        b = context[self.block]
        context_dict = {}
        for d in context.dicts:
            context_dict.update(d)
        # can only take string keys
        for k in context_dict.keys():
            if not isinstance(k, string_types):
                del context_dict[k]
        return b, context_dict


class RenderNode(BaseNode):
    def render(self, context):
        b, context_dict = super(RenderNode, self).prepare_context_data(context)
        return b.render(**context_dict)


@register.tag('render')
def render(parser, token):
    block = token.split_contents()[1:][0]
    return RenderNode(block)


class RenderJSNode(BaseNode):
    def render(self, context):
        b, context_dict = super(
            RenderJSNode, self).prepare_context_data(context)
        return b.render_js(**context_dict)


@register.tag('renderjs')
def renderjs(parser, token):
    block = token.split_contents()[1:][0]
    return RenderJSNode(block)


class RenderCSSNode(BaseNode):
    def render(self, context):
        b, context_dict = super(
            RenderCSSNode, self).prepare_context_data(context)
        return b.render_css(**context_dict)


@register.tag('rendercss')
def rendercss(parser, token):
    block = token.split_contents()[1:][0]
    return RenderCSSNode(block)


class RenderSummaryNode(BaseNode):
    def render(self, context):
        b, context_dict = super(
            RenderSummaryNode, self).prepare_context_data(context)
        return b.render_summary(**context_dict)


@register.tag('rendersummary')
def rendersummary(parser, token):
    block = token.split_contents()[1:][0]
    return RenderSummaryNode(block)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import sys
from django.db import (
    migrations,
    DatabaseError, OperationalError, ProgrammingError
)


class ConditionalDeleteModel(migrations.DeleteModel):
    def database_forwards(
            self, app_label, schema_editor, from_state, to_state):
        # The tests use this model
        if 'runtests.py' in sys.argv or 'test' in sys.argv or \
           'jenkins' in sys.argv:
            return

        try:
            super(ConditionalDeleteModel, self).database_forwards(
                app_label, schema_editor, from_state, to_state)
        except (DatabaseError, OperationalError, ProgrammingError):
            """ if it fails, it's totally fine. it just means that the
            table we wanted to delete doesn't exist. so we ignore it.
            """
            pass


class Migration(migrations.Migration):

    dependencies = [
        ('pagetree', '0001_initial'),
    ]

    operations = [
        ConditionalDeleteModel(
            name='TestBlock',
        ),
    ]



from __future__ import unicode_literals

from django.test import TestCase
from django.contrib.auth.models import User, AnonymousUser
from pagetree.helpers import get_hierarchy, get_section_from_path
from pagetree.helpers import block_submitted
from pagetree.models import Hierarchy


class TestGetHierarchy(TestCase):
    def test_empty(self):
        self.assertEqual(Hierarchy.objects.count(), 0)
        get_hierarchy()
        self.assertEqual(Hierarchy.objects.count(), 1)

    def test_idempotent(self):
        self.assertEqual(Hierarchy.objects.count(), 0)
        get_hierarchy()
        self.assertEqual(Hierarchy.objects.count(), 1)
        get_hierarchy()
        self.assertEqual(Hierarchy.objects.count(), 1)

    def test_default_args(self):
        h = get_hierarchy()
        self.assertEqual(h.name, "main")
        self.assertEqual(h.base_url, "/")

    def test_dont_care_about_type(self):
        h = get_hierarchy()
        h2 = get_hierarchy(h)
        self.assertEqual(h.id, h2.id)


class TestGetSectionFromPath(TestCase):
    def test_empty(self):
        s = get_section_from_path("/")
        self.assertEqual(s.hierarchy.name, "main")


class StubBlockNoSubmit(object):
    pass


class StubBlockSubmit(object):
    def __init__(self, submit_value=False, unlocked_value=False):
        self.submit_value = submit_value
        self.unlocked_value = unlocked_value

    def needs_submit(self):
        return self.submit_value

    def unlocked(self, u):
        return self.unlocked_value


class TestBlockSubmitted(TestCase):
    def test_anonymous(self):
        u = AnonymousUser()
        self.assertFalse(block_submitted(None, u))

    def test_no_submit(self):
        u = User.objects.create(username="test")
        b = StubBlockNoSubmit()
        self.assertTrue(block_submitted(b, u))

    def test_submit_returns_false(self):
        u = User.objects.create(username="test")
        b = StubBlockSubmit(submit_value=False)
        self.assertTrue(block_submitted(b, u))

    def test_submit_unlocked(self):
        u = User.objects.create(username="test")
        b = StubBlockSubmit(submit_value=True, unlocked_value=True)
        self.assertTrue(block_submitted(b, u))

    def test_submit_locked(self):
        u = User.objects.create(username="test")
        b = StubBlockSubmit(submit_value=True, unlocked_value=False)
        self.assertFalse(block_submitted(b, u))



from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.test.testcases import TestCase
from pagetree.models import Hierarchy, Section, UserPageVisit
from pagetree.test_models import TestBlock
import factory


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = User

    username = factory.Sequence(lambda n: "user%03d" % n)
    password = factory.PostGenerationMethodCall('set_password', 'test')


class HierarchyFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Hierarchy

    name = "main"
    base_url = "/pages/"


class RootSectionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Section

    hierarchy = factory.SubFactory(HierarchyFactory)
    label = "Root"
    path = "path"
    slug = "slug"
    depth = 0


class TestBlockFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = TestBlock

    body = "Test block body"


class ModuleFactory(object):
    def __init__(self, hname='main', base_url='/pages/'):
        hierarchy = HierarchyFactory(name=hname, base_url=base_url)
        root = hierarchy.get_root()
        root.add_child_section_from_dict(
            {'label': "One", 'slug': "one",
             'children': [{'label': "Three", 'slug': "introduction"}]})
        root.add_child_section_from_dict({'label': "Two", 'slug': "two"})

        blocks = [{'label': 'Welcome to your new Forest Site',
                   'css_extra': '',
                   'block_type': 'Test Block',
                   'body': 'You should now use the edit link to add content'}]
        root.add_child_section_from_dict({'label': 'Four', 'slug': 'four',
                                          'pageblocks': blocks})

        self.root = root


class PagetreeTestCase(TestCase):
    def setUp(self):
        super(PagetreeTestCase, self).setUp()

        ModuleFactory("one", "/pages/one/")
        ModuleFactory("two", "/pages/two/")

        self.hierarchy_one = Hierarchy.objects.get(name='one')
        self.hierarchy_two = Hierarchy.objects.get(name='two')


class UserPageVisitFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = UserPageVisit

    user = factory.SubFactory(UserFactory)
    section = factory.SubFactory(RootSectionFactory)



from __future__ import unicode_literals

from django.test import TestCase
from pagetree.tests.factories import RootSectionFactory, UserFactory

from pagetree.templatetags.user_status import is_section_unlocked


class IsSectionUnlockedTest(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.section = RootSectionFactory()

    def test_is_section_unlocked_empty_section(self):
        self.assertEqual(
            is_section_unlocked(self.section, self.user), '1')
